@@ -4,9 +4,7 @@ use strict;
 use warnings;
 
 use Carp ();
-use Data::OptList ();
-use Params::Util ();
-use Sub::Install 0.92 ();
+use Sub::Install;
 
 =head1 NAME
 
@@ -14,11 +12,13 @@ Sub::Exporter - a sophisticated exporter for custom-built routines
 
 =head1 VERSION
 
-version 0.977_02
+version 0.90
+
+  $Id: /my/cs/projects/export/trunk/lib/Sub/Exporter.pm 19737 2006-03-13T16:24:13.590498Z rjbs  $
 
 =cut
 
-our $VERSION = '0.977_02';
+our $VERSION = '0.90';
 
 =head1 SYNOPSIS
 
@@ -114,7 +114,7 @@ subroutines:
 
   sub _curry_class {
     my ($class, $name) = @_;
-    sub { $class->$name(@_); };
+    sub { $class->$name(@_);
   }
 
 Because of the way that exporters and Sub::Exporter work, any package that
@@ -174,48 +174,27 @@ The following keys are valid in C<%config>:
   exports - a list of routines to provide for exporting; each routine may be
             followed by generator
   groups  - a list of groups to provide for exporting; each must be followed by
-            either (a) a list of exports, possibly with arguments for each
-            export, or (b) a generator
-
+            a list of exports, possibly with arguments for each export
   collectors - a list of names into which values are collected for use in
                routine generation; each name may be followed by a validator
 
-In addition to the basic options above, a few more advanced options may be
-passed:
-
-  into_level - how far up the caller stack to look for a target (default 0)
-  into       - an explicit target (package) into which to export routines
-
-In other words: Sub::Exporter installs a C<import> routine which, when called,
-exports routines to the calling namespace.  The C<into> and C<into_level>
-options change where those exported routines are installed.
-
-  generator  - a callback used to produce the code that will be installed
-               default: Sub::Exporter::default_generator
-
-  installer  - a callback used to install the code produced by the generator
-               default: Sub::Exporter::default_installer
-
-For information on how these callbacks are used, see the documentation for
-C<L</default_generator>> and C<L</default_installer>>.
-
-=head2 Export Configuration
+=head2 C<exports> Configuration
 
 The C<exports> list may be provided as an array reference or a hash reference.
 The list is processed in such a way that the following are equivalent:
 
-  { exports => [ qw(foo bar baz), quux => \&quux_generator ] }
+  { exports => [ qw(foo bar baz), quux => \&quuz_generator ] }
 
   { exports =>
-    { foo => undef, bar => undef, baz => undef, quux => \&quux_generator } }
+    { foo => undef, bar => undef, baz => undef, quux => \&quuz_generator } }
 
-Generators are code that return coderefs.  They are called with four
+Generators are coderefs that return coderefs.  They are called with four
 parameters:
 
   $class - the class whose exporter has been called (the exporting class)
   $name  - the name of the export for which the routine is being build
  \%arg   - the arguments passed for this export
- \%col   - the collections for this import
+ \%coll  - the collections for this import
 
 Given the configuration in the L</SYNOPSIS>, the following C<use> statement:
 
@@ -232,15 +211,10 @@ would result in the following call to C<&build_reformatter>:
     { defaults => { eol => 'CR' } },
   );
 
-The returned coderef (C<$code>) would then be installed as C<make_narrow> in the
+The returned coderef (<$code>) would then be installed as C<make_narrow> in the
 calling package.
 
-Instead of providing a coderef in the configuration, a reference to a method
-name may be provided.  This method will then be called on the invocant of the
-C<import> method.  (In this case, we do not pass the C<$class> parameter, as it
-would be redundant.)
-
-=head2 Group Configuration
+=head2 C<groups> Configuration
 
 The C<groups> list can be passed in the same forms as C<exports>.  Groups must
 have values to be meaningful, which may either list exports that make up the
@@ -296,7 +270,7 @@ configured, it will be empty, and nothing will happen.
 Another group is also created if not defined: C<all>.  The C<all> group
 contains all the exports from the exports list.
 
-=head2 Collector Configuration
+=head2 C<collectors> Configuration
 
 The C<collectors> entry in the exporter configuration gives names which, when
 found in the import call, have their values collected and passed to every
@@ -337,18 +311,6 @@ default easily:
 
   collectors => { defaults => sub { return (exists $_[0]->{data}) ? 0 : 1 } }
 
-Collector coderefs can also be used as hooks to perform arbitrary actions
-before anything is exported.
-
-When the coderef is called, it is passed the value of the collection and a
-hashref containing the following entries:
-
-  name        - the name of the collector
-  config      - the exporter configuration (hashref)
-  import_args - the arguments passed to the exporter, sans collections (aref)
-  class       - the package on which the importer was called
-  into        - the package into which exports will be exported
-
 =head1 CALLING THE EXPORTER
 
 Arguments to the exporter (that is, the arguments after the module name in a
@@ -376,30 +338,76 @@ however, contain a reference to a scalar.  If that is the case, a reference the
 generated routine will be placed in the scalar referenced by C<-as>.  It will
 not be installed into the calling package.
 
-=head2 Special Exporter Arguments
-
-The generated exporter accept some special options, which may be passed as the
-first argument, in a hashref.
-
-These options are:
-
-  into_level
-  into
-  generator
-  installer
-
-These override the same-named configuration options described in L</EXPORTER
-CONFIGURATION>.
-
 =cut
 
+# This produces an array of arrays; the inner arrays are name/value pairs.
+# Values will be either "undef" or a reference.  $must_be is either a scalar or
+# array of scalars; it defines what kind(s) of refs may be values.  If an
+# invalid value is found, an exception is thrown.
+# possible inputs:
+#  undef    -> []
+#  hashref  -> [ [ key1 => value1 ] ... ] # non-ref values become undef
+#  arrayref -> every value followed by a ref becomes a pair: [ value => ref   ]
+#              every value followed by undef becomes a pair: [ value => undef ]
+#              otherwise, it becomes [ value => undef ] like so:
+#              [ "a", "b", [ 1, 2 ] ] -> [ [ a => undef ], [ b => [ 1, 2 ] ] ]
+sub _canonicalize_opt_list {
+  my ($opt_list, $require_unique, $must_be) = @_;
+
+  return [] unless $opt_list;
+
+  $opt_list = [
+    map { $_ => (ref $opt_list->{$_} ? $opt_list->{$_} : ()) } keys %$opt_list
+  ] if ref $opt_list eq 'HASH';
+
+  my @return;
+
+  for (my $i = 0; $i < @$opt_list; $i++) {
+    my $name = $opt_list->[$i];
+    my $value;
+
+    if    ($i == $#$opt_list)             { $value = undef;            }
+    elsif (not defined $opt_list->[$i+1]) { $value = undef; $i++       }
+    elsif (ref $opt_list->[$i+1])         { $value = $opt_list->[++$i] }
+    else                                  { $value = undef;            }
+
+    if ($must_be and defined $value) {
+      $must_be = [ $must_be ] unless ref $must_be;
+      my $ref = ref $value;
+      Carp::croak "$ref-ref values are not allowed"
+        unless grep { $ref eq $_ } @$must_be;
+    }
+
+    push @return, [ $name => $value ];
+  }
+
+  if ($require_unique) {
+    my %seen;
+    map { Carp::croak "multiple definitions provided for $_" if $seen{$_}++ }
+    map { $_->[0] } @return;
+  }
+
+  return \@return;
+}
+
+# This turns a canonicalized opt_list (see above) into a hash.
+sub _expand_opt_list {
+  my ($opt_list, $must_be) = @_;
+  return {} unless $opt_list;
+  return $opt_list if ref $opt_list eq 'HASH';
+
+  $opt_list = _canonicalize_opt_list($opt_list, 1, $must_be);
+  my %hash = map { $_->[0] => $_->[1] } @$opt_list;
+  return \%hash;
+}
+
 # Given a potential import name, this returns the group name -- if it's got a
 # group prefix.
 sub _group_name {
   my ($name) = @_;
 
-  return if (index q{-:}, (substr $name, 0, 1)) == -1;
-  return substr $name, 1;
+  return unless $name =~ s/\A[-:]//;
+  return $name;
 }
 
 # \@groups is a canonicalized opt list of exports and groups this returns
@@ -410,35 +418,31 @@ sub _expand_groups {
   my ($class, $config, $groups, $collection, $seen, $merge) = @_;
   $seen  ||= {};
   $merge ||= {};
+
   my @groups = @$groups;
 
-  for my $i (reverse 0 .. $#groups) {
+  for my $i (reverse 0 .. $#$groups) {
+    # this isn't a group, let it be
     if (my $group_name = _group_name($groups[$i][0])) {
-      my $seen = { %$seen }; # faux-dynamic scoping
+      # we already dealt with this, remove it
+      if ($seen->{ $group_name }) {
+        splice @groups, $i, 1;
+        next;
+      }
 
+      # rewrite the group
       splice @groups, $i, 1,
         _expand_group($class, $config, $groups[$i], $collection, $seen, $merge);
     } else {
-      # there's nothing to munge in this export's args
       next unless my %merge = %$merge;
-
-      # we have things to merge in; do so
       my $prefix = (delete $merge{-prefix}) || '';
       my $suffix = (delete $merge{-suffix}) || '';
-
-      if (
-        Params::Util::_CODELIKE($groups[$i][1]) ## no critic Private
-        or
-        Params::Util::_SCALAR0($groups[$i][1]) ## no critic Private
-      ) {
-        # this entry was build by a group generator
-        $groups[$i][0] = $prefix . $groups[$i][0] . $suffix;
+      if (ref $groups[$i][1] eq 'CODE') {
+         $groups[$i][0] = $prefix . $groups[$i][0] . $suffix;
       } else {
-        my $as
-          = ref $groups[$i][1]{-as} ? $groups[$i][1]{-as}
-          :     $groups[$i][1]{-as} ? $prefix . $groups[$i][1]{-as} . $suffix
-          :                           $prefix . $groups[$i][0]      . $suffix;
-
+        my $as = ref $groups[$i][1]{-as}
+          ? $groups[$i][1]{-as}
+          : $prefix . ($groups[$i][1]{-as}||$groups[$i][0]) . $suffix;
         $groups[$i][1] = { %{ $groups[$i][1] }, %merge, -as => $as };
       }
     }
@@ -455,11 +459,6 @@ sub _expand_group {
   my ($group_name, $group_arg) = @$group;
   $group_name = _group_name($group_name);
 
-  Carp::croak qq(group "$group_name" is not exported by the $class module)
-    unless exists $config->{groups}{$group_name};
-
-  return if $seen->{$group_name}++;
-
   if (ref $group_arg) {
     my $prefix = (delete $merge->{-prefix}||'') . ($group_arg->{-prefix}||'');
     my $suffix = ($group_arg->{-suffix}||'') . (delete $merge->{-suffix}||'');
@@ -471,34 +470,23 @@ sub _expand_group {
     };
   }
 
+  Carp::croak qq(group "$group_name" is not exported by the $class module)
+    unless exists $config->{groups}{$group_name};
+
+  $seen->{$group_name} = 1;
+  
   my $exports = $config->{groups}{$group_name};
 
-  if (
-    Params::Util::_CODELIKE($exports) ## no critic Private
-    or
-    Params::Util::_SCALAR0($exports) ## no critic Private
-  ) {
-    # I'm not very happy with this code for hiding -prefix and -suffix, but
-    # it's needed, and I'm not sure, offhand, how to make it better.
-    # -- rjbs, 2006-12-05
-    my $group_arg = $group_arg ? { %$group_arg } : {};
-    delete $group_arg->{-prefix};
-    delete $group_arg->{-suffix};
-
-    my $group = Params::Util::_CODELIKE($exports) ## no critic Private
-              ? $exports->($class, $group_name, $group_arg, $collection)
-              : $class->$$exports($group_name, $group_arg, $collection);
-
+  if (ref $exports eq 'CODE') {
+    my $group = $exports->($class, $group_name, $group_arg, $collection);
     Carp::croak qq(group generator "$group_name" did not return a hashref)
       if ref $group ne 'HASH';
-
     my $stuff = [ map { [ $_ => $group->{$_} ] } keys %$group ];
     return @{
       _expand_groups($class, $config, $stuff, $collection, $seen, $merge)
     };
   } else {
-    $exports
-      = Data::OptList::mkopt($exports, "$group_name exports");
+    $exports = _canonicalize_opt_list($exports);
 
     return @{
       _expand_groups($class, $config, $exports, $collection, $seen, $merge)
@@ -509,39 +497,23 @@ sub _expand_group {
 # Given a config and pre-canonicalized importer args, remove collections from
 # the args and return them.
 sub _collect_collections {
-  my ($config, $import_args, $class, $into) = @_;
+  my ($config, $import_args) = @_;
   my %collection;
 
-  my @collections
-    = map  { splice @$import_args, $_, 1 }
-      grep { exists $config->{collectors}{ $import_args->[$_][0] } }
-      reverse 0 .. $#$import_args;
+  for my $collection (keys %{ $config->{collectors} }) {
+    next unless my @indexes
+      = grep { $import_args->[$_][0] eq $collection } (0 .. $#$import_args);
 
-  my %seen;
-  for my $collection (@collections) {
-    my ($name, $value) = @$collection;
+    Carp::croak "collection $collection provided multiple times in import"
+      if @indexes > 1;
 
-    Carp::croak "collection $name provided multiple times in import"
-      if $seen{ $name }++;
+    my $value = splice @$import_args, $indexes[0], 1;
+    $collection{ $collection } = $value->[1];
 
-    if (ref(my $hook = $config->{collectors}{$name})) {
-      my $arg = {
-        name        => $name,
-        config      => $config,
-        import_args => $import_args,
-        class       => $class,
-        into        => $into,
-      };
-
-      my $error_msg = "collection $name failed validation";
-      if (Params::Util::_SCALAR0($hook)) { ## no critic Private
-        Carp::croak $error_msg unless $class->$$hook($value, $arg);
-      } else {
-        Carp::croak $error_msg unless $hook->($value, $arg);
-      }
+    if (ref(my $validator = $config->{collectors}{$collection})) {
+      Carp::croak "collection $collection failed validation"
+        unless $validator->($collection{$collection});
     }
-
-    $collection{ $name } = $value;
   }
 
   return \%collection;
@@ -549,51 +521,41 @@ sub _collect_collections {
 
 =head1 SUBROUTINES
 
-=head2 setup_exporter
+=head2 C< setup_exporter >
 
 This routine builds and installs an C<import> routine.  It is called with one
 argument, a hashref containing the exporter configuration.  Using this, it
 builds an exporter and installs it into the calling package with the name
-"import."  In addition to the normal exporter configuration, a few named
+"import."  In addition to the normal exporter configuration, two named
 arguments may be passed in the hashref:
 
-  into       - into what package should the exporter be installed
-  into_level - into what level up the stack should the exporter be installed
-  as         - what name should the installed exporter be given
-
-By default the exporter is installed with the name C<import> into the immediate
-caller of C<setup_exporter>.  In other words, if your package calls
-C<setup_exporter> without providing any of the three above arguments, it will
-have an C<import> routine installed.
-
-Providing both C<into> and C<into_level> will cause an exception to be thrown.
+  into - into what package should the exporter be installed (defaults to caller)
+  as   - what name should the installed exporter be given (defaults to "import")
 
 The exporter is built by C<L</build_exporter>>.
 
 =cut
 
+# \%special is for experimental options that may or may not be kept around and,
+# probably, moved to \%config.  These are also passed along to build_exporter.
+
 sub setup_exporter {
-  my ($config)  = @_;
-
-  Carp::croak q(into and into_level may not both be supplied to exporter)
-    if exists $config->{into} and exists $config->{into_level};
+  my ($config, $special)  = @_;
+  $special ||= {};
 
+  my $into = delete $config->{into} || caller(0);
   my $as   = delete $config->{as}   || 'import';
-  my $into
-    = exists $config->{into}       ? delete $config->{into}
-    : exists $config->{into_level} ? caller(delete $config->{into_level})
-    :                                caller(0);
 
-  my $import = build_exporter($config);
+  my $import = build_exporter($config, $special);
 
-  Sub::Install::reinstall_sub({
+  Sub::Install::install_sub({
     code => $import,
     into => $into,
     as   => $as,
   });
 }
 
-=head2 build_exporter
+=head2 C< build_exporter >
 
 Given a standard exporter configuration, this routine builds and returns an
 exporter -- that is, a subroutine that can be installed as a class method to
@@ -604,65 +566,18 @@ the exporter as a package's import routine.
 
 =cut
 
-sub _key_intersection {
-  my ($x, $y) = @_;
-  my %seen = map { $_ => 1 } keys %$x;
-  my @names = grep { $seen{$_} } keys %$y;
-}
+sub build_exporter {
+  my ($config, $special) = @_;
+  $special ||= {};
 
-# Given the config passed to setup_exporter, which contains sugary opt list
-# data, rewrite the opt lists into hashes, catch a few kinds of invalid
-# configurations, and set up defaults.  Since the config is a reference, it's
-# rewritten in place.
-my %valid_config_key;
-BEGIN {
-  %valid_config_key =
-    map { $_ => 1 }
-    qw(collectors installer generator exports groups into into_level),
-    qw(exporter), # deprecated
-}
+  # this option name, if nothing else, needs to be improved before it is
+  # accepted as a core feature -- rjbs, 2006-03-09
+  $special->{export} ||= \&_export;
+  
+  $config->{$_} = _expand_opt_list($config->{$_}, 'CODE')
+    for qw(exports collectors);
 
-sub _rewrite_build_config {
-  my ($config) = @_;
-
-  if (my @keys = grep { not exists $valid_config_key{$_} } keys %$config) {
-    Carp::croak "unknown options (@keys) passed to Sub::Exporter";
-  }
-
-  Carp::croak q(into and into_level may not both be supplied to exporter)
-    if exists $config->{into} and exists $config->{into_level};
-
-  # XXX: Remove after deprecation period.
-  if ($config->{exporter}) {
-    Carp::cluck "'exporter' argument to build_exporter is deprecated. Use 'installer' instead; the semantics are identical.";
-    $config->{installer} = delete $config->{exporter};
-  }
-
-  Carp::croak q(into and into_level may not both be supplied to exporter)
-    if exists $config->{into} and exists $config->{into_level};
-
-  for (qw(exports collectors)) {
-    $config->{$_} = Data::OptList::mkopt_hash(
-      $config->{$_},
-      $_,
-      [ 'CODE', 'SCALAR' ],
-    );
-  }
-
-  if (my @names = _key_intersection(@$config{qw(exports collectors)})) {
-    Carp::croak "names (@names) used in both collections and exports";
-  }
-
-  $config->{groups} = Data::OptList::mkopt_hash(
-      $config->{groups},
-      'groups',
-      [
-        'HASH',   # standard opt list
-        'ARRAY',  # standard opt list
-        'CODE',   # group generator
-        'SCALAR', # name of group generation method
-      ]
-    );
+  $config->{groups} = _expand_opt_list($config->{groups}, [ 'HASH', 'CODE' ]);
 
   # by default, export nothing
   $config->{groups}{default} ||= [];
@@ -670,246 +585,101 @@ sub _rewrite_build_config {
   # by default, build an all-inclusive 'all' group
   $config->{groups}{all} ||= [ keys %{ $config->{exports} } ];
 
-  $config->{generator} ||= \&default_generator;
-  $config->{installer} ||= \&default_installer;
-}
-
-sub build_exporter {
-  my ($config) = @_;
-
-  _rewrite_build_config($config);
-
   my $import = sub {
     my ($class) = shift;
-
-    # XXX: clean this up -- rjbs, 2006-03-16
-    my $special = (ref $_[0]) ? shift(@_) : {};
-    Carp::croak q(into and into_level may not both be supplied to exporter)
-      if exists $special->{into} and exists $special->{into_level};
-
-    if ($special->{exporter}) {
-      Carp::cluck "'exporter' special import argument is deprecated. Use 'installer' instead; the semantics are identical.";
-      $special->{installer} = delete $special->{exporter};
-    }
-
-    my $into
-      = defined $special->{into}       ? delete $special->{into}
-      : defined $special->{into_level} ? caller(delete $special->{into_level})
-      : defined $config->{into}        ? $config->{into}
-      : defined $config->{into_level}  ? caller($config->{into_level})
-      :                                  caller(0);
-
-    my $generator = delete $special->{generator} || $config->{generator};
-    my $installer = delete $special->{installer} || $config->{installer};
+    my ($into)  = caller(0);
 
     # this builds a AOA, where the inner arrays are [ name => value_ref ]
-    my $import_args = Data::OptList::mkopt([ @_ ]);
-
-    # is this right?  defaults first or collectors first? -- rjbs, 2006-06-24
-    $import_args = [ [ -default => undef ] ] unless @$import_args;
-
-    my $collection = _collect_collections($config, $import_args, $class, $into);
+    my $import_args = _canonicalize_opt_list([ @_ ]);
+    
+    my $collection = _collect_collections($config, $import_args);
 
+    $import_args = [ [ -default => 1 ] ] unless @$import_args;
     my $to_import = _expand_groups($class, $config, $import_args, $collection);
 
     # now, finally $import_arg is really the "to do" list
-    _do_import(
-      {
-        class     => $class,
-        col       => $collection,
-        config    => $config,
-        into      => $into,
-        generator => $generator,
-        installer => $installer,
-      },
-      $to_import,
-    );
+    for (@$to_import) {
+      my ($name, $arg) = @$_;
+
+      my ($generator, $as);
+
+      if ($arg and ref $arg eq 'CODE') {
+        $generator = sub { $arg };
+        $as = $name;
+      } else {
+        $arg = { $arg ? %$arg : () };
+
+        Carp::croak qq("$name" is not exported by the $class module)
+          unless (exists $config->{exports}{$name});
+
+        $generator = $config->{exports}{$name};
+
+        $as = exists $arg->{-as} ? (delete $arg->{-as}) : $name;
+      }
+
+      $special->{export}->(
+        $class, $generator, $name, $arg, $collection, $as, $into
+      );
+    }
   };
 
   return $import;
 }
 
-sub _do_import {
-  my ($arg, $to_import) = @_;
-
-  my @todo;
-
-  for my $pair (@$to_import) {
-    my ($name, $import_arg) = @$pair;
-
-    my ($generator, $as);
-
-    if ($import_arg and Params::Util::_CODELIKE($import_arg)) { ## no critic
-      # This is the case when a group generator has inserted name/code pairs.
-      $generator = sub { $import_arg };
-      $as = $name;
-    } else {
-      $import_arg = { $import_arg ? %$import_arg : () };
-
-      Carp::croak qq("$name" is not exported by the $arg->{class} module)
-        unless exists $arg->{config}{exports}{$name};
-
-      $generator = $arg->{config}{exports}{$name};
-
-      $as = exists $import_arg->{-as} ? (delete $import_arg->{-as}) : $name;
-    }
-
-    my $code = $arg->{generator}->(
-      { 
-        class     => $arg->{class},
-        name      => $name,
-        arg       => $import_arg,
-        col       => $arg->{col},
-        generator => $generator,
-      }
-    );
-
-    push @todo, $as, $code;
-  }
-
-  $arg->{installer}->(
-    {
-      class => $arg->{class},
-      into  => $arg->{into},
-      col   => $arg->{col},
-    },
-    \@todo,
+# the default installer; it does what Sub::Exporter promises: call generators
+# with the three normal arguments, then install the code into the target
+# package
+sub _export {
+  my ($class, $generator, $name, $arg, $collection, $as, $into) = @_;
+  _install(
+    _generate($class, $generator, $name, $arg, $collection, $as, $into),
+    $into,
+    $as,
   );
 }
 
-## Cute idea, possibly for future use: also supply an "unimport" for:
-## no Module::Whatever qw(arg arg arg);
-# sub _unexport {
-#   my (undef, undef, undef, undef, undef, $as, $into) = @_;
-# 
-#   if (ref $as eq 'SCALAR') {
-#     undef $$as;
-#   } elsif (ref $as) {
-#     Carp::croak "invalid reference type for $as: " . ref $as;
-#   } else {
-#     no strict 'refs';
-#     delete &{$into . '::' . $as};
-#   }
-# }
+sub _generate {
+  my ($class, $generator, $name, $arg, $collection, $as, $into) = @_;
 
-=head2 default_generator
-
-This is Sub::Exporter's default generator.  It takes bits of configuration that
-have been gathered during the import and turns them into a coderef that can be
-installed.
-
-  my $code = default_generator(\%arg);
-
-Passed arguments are:
-
-  class - the class on which the import method was called
-  name  - the name of the export being generated
-  arg   - the arguments to the generator
-  col   - the collections
-
-  generator - the generator to be used to build the export (code or scalar ref)
-
-=cut
-
-sub default_generator {
-  my ($arg) = @_;
-  my ($class, $name, $generator) = @$arg{qw(class name generator)};
-
-  if (not defined $generator) {
-    my $code = $class->can($name)
-      or Carp::croak "can't locate exported subroutine $name via $class";
-    return $code;
-  }
-
-  # I considered making this "$class->$generator(" but it seems that
-  # overloading precedence would turn an overloaded-as-code generator object
-  # into a string before code. -- rjbs, 2006-06-11
-  return $generator->($class, $name, $arg->{arg}, $arg->{col})
-    if Params::Util::_CODELIKE($generator); ## no critic Private
-
-  # This "must" be a scalar reference, to a generator method name.
-  # -- rjbs, 2006-12-05
-  return $class->$$generator($name, $arg->{arg}, $arg->{col});
+  my $code = $generator
+           ? $generator->($class, $name, $arg, $collection)
+           : $class->can($name); 
 }
 
-=head2 default_installer
+sub _install {
+  my ($code, $into, $as) = @_;
+  # Allow as isa ARRAY to push onto an array?
+  # Allow into isa HASH to install name=>code into hash?
 
-This is Sub::Exporter's default installer.  It does what Sub::Exporter
-promises: it installs code into the target package.
-
-  default_installer(\%arg, \@to_export);
-
-Passed arguments are:
-
-  into - the package into which exports should be delivered
-
-C<@to_export> is a list of name/value pairs.  The default exporter assigns code
-(the values) to named slots (the names) in the given package.  If the name is a
-scalar reference, the scalar reference is made to point to the code reference
-instead.
-
-=cut
-
-sub default_installer {
-  my ($arg, $to_export) = @_;
-
-  for (my $i = 0; $i < @$to_export; $i += 2) {
-    my ($as, $code) = @$to_export[ $i, $i+1 ];
-
-    # Allow as isa ARRAY to push onto an array?
-    # Allow into isa HASH to install name=>code into hash?
-
-    if (ref $as eq 'SCALAR') {
-      $$as = $code;
-    } elsif (ref $as) {
-      Carp::croak "invalid reference type for $as: " . ref $as;
-    } else {
-      Sub::Install::reinstall_sub({
-        code => $code,
-        into => $arg->{into},
-        as   => $as
-      });
-    }
+  if (ref $as eq 'SCALAR') {
+    $$as = $code;
+  } elsif (ref $as) {
+    Carp::croak "invalid reference type for $as: " . ref $as;
+  } else {
+    Sub::Install::install_sub({ code => $code, into => $into, as => $as });
   }
 }
 
-sub default_exporter {
-  Carp::croak "default_exporter is deprecated; call default_installer instead; the semantics are identical";
-  goto &default_installer;
-}
-
 =head1 EXPORTS
 
 Sub::Exporter also offers its own exports: the C<setup_exporter> and
 C<build_exporter> routines described above.  It also provides a special "setup"
-group, which will setup an exporter using the parameters passed to it.
+group, which will setup an exporter using the parameters passed to it.  This
+group can be passed a list of subroutines names to export, instead of the
+normal configuration hash.
 
 =cut
 
 setup_exporter({
   exports => [
     qw(setup_exporter build_exporter),
-    _import => sub { build_exporter($_[2]) },
+    _import => sub { splice @_, 0, 2; build_exporter(@_) },
   ],
   groups  => {
     all   => [ qw(setup_exporter build_export) ],
-  },
-  collectors => { -setup => \&_setup },
-});
-
-sub _setup {
-  my ($value, $arg) = @_;
-
-  if (ref $value eq 'HASH') {
-    push @{ $arg->{import_args} }, [ _import => { -as => 'import', %$value } ];
-    return 1;
-  } elsif (ref $value eq 'ARRAY') {
-    push @{ $arg->{import_args} },
-      [ _import => { -as => 'import', exports => $value } ];
-    return 1;
+    setup => { _import => { -as => 'import' } }
   }
-  return;
-}
+});
 
 =head1 COMPARISONS
 
@@ -941,8 +711,8 @@ complex uses, Sub::Exporter makes hard things possible, which would not be
 possible with Exporter. 
 
 When using a module that uses Sub::Exporter, users familiar with Exporter will
-probably see no difference in the basics.  These two lines do about the same
-thing in whether the exporting module uses Exporter or Sub::Exporter.
+probably see difference in the basics.  These two lines do about the same thing
+in whether the exporting module uses Exporter or Sub::Exporter.
 
   use Some::Module qw(foo bar baz);
   use Some::Module qw(foo :bar baz);
@@ -994,8 +764,7 @@ variables, at runtime.)
 L<Perl6::Export> isn't actually attribute based, but looks similar.  Its syntax
 is borrowed from Perl 6, and implemented by a source filter.  It is a prototype
 of an interface that is still being designed.  It should probably be avoided
-for production work.  On the other hand, L<Perl6::Export::Attrs> implements
-Perl 6-like exporting, but translates it into Perl 5 by providing attributes.
+for production work.
 
 =item * Other Exporters
 
@@ -1017,13 +786,14 @@ variables for its configuration.
 
 =cut
 
+# This would be cool:
+# use Food qr/\Aartificial/ => { -prefix => 'non_' };
+
 =over
 
 =item * write a set of longer, more demonstrative examples
 
-=item * solidify the "custom exporter" interface (see C<&default_exporter>)
-
-=item * add an "always" group
+=item * solidify the "custom build and install" interface (see C<&_export>)
 
 =back
 
@@ -1035,20 +805,18 @@ Ricardo SIGNES, C<< <rjbs@cpan.org> >>
 
 Hans Dieter Pearcey provided helpful advice while I was writing Sub::Exporter.
 Ian Langworth and Shawn Sorichetti asked some good questions and hepled me
-improve my documentation quite a bit.  Yuval Kogman helped me find a bunch of
-little problems.
-
-Thanks, guys! 
+improve my documentation quite a bit.  Thanks, guys!
 
 =head1 BUGS
 
-Please report any bugs or feature requests through the web interface at
-L<http://rt.cpan.org>. I will be notified, and then you'll automatically be
-notified of progress on your bug as I make changes.
+Please report any bugs or feature requests to C<bug-sub-exporter@rt.cpan.org>,
+or through the web interface at L<http://rt.cpan.org>. I will be notified, and
+then you'll automatically be notified of progress on your bug as I make
+changes.
 
 =head1 COPYRIGHT
 
-Copyright 2006-2007, Ricardo SIGNES.  This program is free software;  you can
+Copyright 2006 Ricardo SIGNES.  This program is free software;  you can
 redistribute it and/or modify it under the same terms as Perl itself.
 
 =cut
